# %% imports
from autogaita.resources.utils import (
    write_issues_to_textfile,
    bin_num_to_percentages,
    compute_angle,
    write_angle_warning,
)
import os
import pandas as pd
import numpy as np
import warnings

# %% constants
from autogaita.resources.constants import SC_PERCENTAGE_COL
from autogaita.universal3D.universal3D_constants import (
    LEGS,
    LEGS_COLFORMAT,
    OUTPUTS,
    ORIGINAL_XLS_FILENAME,
    Y_STANDARDISED_XLS_FILENAME,
    NORMALISED_XLS_FILENAME,
    AVERAGE_XLS_FILENAME,
    STD_XLS_FILENAME,
    SEPARATOR_IDX,
    LEG_COL,
    EXCLUDED_COLS_IN_AV_STD_DFS,
    REORDER_COLS_IN_STEP_NORMDATA,
)

# %% workflow step #3 - y-flipping, y-stand, features, df-creation & exports
# Note
# ----
# There is quite a lot going on in this function. We:
# 1) loop through all step cycles for one leg at a time and extract individual SCs' data
# 2) for each step's data we normalise all Z (height) values to the mindfoots minimum
#    if wanted
# 3) we then we flip y columns if needed (to simulate equal walking direction)
# 4) immediately after 3 & 4 and for each step's data separately, we compute and add
#    features (angles, velocities, accelerations)
#    ==> see standardise_y_z_flip_gait_add_features_to_one_step & helper functions a
# 5) immediately after adding features, we normalise a step to bin_num
#    ==> see normalise_one_steps_data & helper functions b
# 6) we add original and normalised steps to all_steps_data and normalised_steps_data
# 7) once we are done with this for a given leg we create average and std dataframes
# 8) we combine legs and store those dfs in a third idx of our dataframe lists
#    ==> see helper functions c for #6 & #7
# 9) we finally output all df-lists in a results dict and export each df-list
#    (of left, right and both legs) in excel files with 3 sheets each
#   ==> see helper functions d


# ...............................  main function  ......................................
def analyse_and_export_stepcycles(data, all_cycles, global_Y_max, info, cfg):
    """Export original-length and normalised XLS files of extracted steps"""
    # ..............................  preparation  .....................................
    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    bin_num = cfg["bin_num"]
    analyse_average_y = cfg["analyse_average_y"]
    flip_gait_direction = cfg["flip_gait_direction"]
    standardise_y_coordinates = cfg["standardise_y_coordinates"]
    # do everything on a copy of the data df
    data_copy = data.copy()
    # for exports, we don't need all_cycles to be separated for runs
    # ==> transform to a list of all SCs for each leg
    all_cycles = flatten_all_cycles(all_cycles)
    # check if there are excel files from a previous run, if so - delete
    delete_previous_xlsfiles(name, results_dir)
    # initialise list of dfs & results
    all_steps_data = [pd.DataFrame(data=None)] * len(OUTPUTS)
    y_standardised_steps_data = [pd.DataFrame(data=None)] * len(OUTPUTS)
    normalised_steps_data = [pd.DataFrame(data=None)] * len(OUTPUTS)
    average_data = [pd.DataFrame(data=None)] * len(OUTPUTS)
    std_data = [pd.DataFrame(data=None)] * len(OUTPUTS)
    sc_num = [None for i in range(len(OUTPUTS))]
    results = {"left": {}, "right": {}, "both": {}}
    # quick warning if cfg is set to not flip gait direction but to standardise y
    if not flip_gait_direction and standardise_y_coordinates:
        message = (
            "\n***********\n! WARNING !\n***********\n"
            + "You are standardising Y coordinates without standardising the direction "
            + "of gait (e.g. all walking from right to left)."
            + "\nThis is correct if you are doing things like treadmill walking but "
            + "can lead to unexpected behaviour otherwise!"
            + "\nMake sure you know what you are doing!"
        )
        print(message)
        write_issues_to_textfile(message, info)
    # .................  loop over legs and populate dfs  ..............................
    # for each step:
    # 1) extract it from data_copy (this_step)
    # 2) normalise z if wanted
    # 3) flip y columns if wanted
    # 4) standardise y to a joint if wanted
    # 5) add features
    # 6) normalise its length to bin_num (this_normalised_step)
    # 7) add this_step to all_steps_data, this_normalised_step to
    #    normalised_steps_data and this_y_standardised_step to y_standardised_steps_data
    for l_idx, legname in enumerate(LEGS):
        # NOTE
        # ----
        # normalised_steps_data is created using y_standardised_step or first_step,
        # leading to average/std dfs being based on y-standardised or raw data
        # automatically
        # 1 step only (highly unlikely in humans)
        if len(all_cycles[legname]) == 1:
            this_step = data_copy.loc[
                all_cycles[legname][0][0] : all_cycles[legname][0][1]
            ]
            if standardise_y_coordinates:
                this_step, y_standardised_step = (
                    standardise_y_z_flip_gait_add_features_to_one_step(
                        this_step, global_Y_max, info, cfg
                    )
                )
                all_steps_data[l_idx] = this_step
                y_standardised_steps_data[l_idx] = y_standardised_step
                normalised_steps_data[l_idx] = normalise_one_steps_data(
                    y_standardised_step, bin_num  # normalising y-standardised here!
                )
            else:
                this_step = standardise_y_z_flip_gait_add_features_to_one_step(
                    this_step, global_Y_max, info, cfg
                )
                all_steps_data[l_idx] = this_step
                normalised_steps_data[l_idx] = normalise_one_steps_data(
                    this_step, bin_num
                )
            sc_num[l_idx] = 1
        # 2 or more steps - build dataframes
        elif len(all_cycles[legname]) > 1:
            # first, add the first step
            first_step = data_copy.loc[
                all_cycles[legname][0][0] : all_cycles[legname][0][1]
            ]
            if standardise_y_coordinates:
                first_step, y_standardised_first_step = (
                    standardise_y_z_flip_gait_add_features_to_one_step(
                        first_step, global_Y_max, info, cfg
                    )
                )
                all_steps_data[l_idx] = first_step
                y_standardised_steps_data[l_idx] = y_standardised_first_step
                normalised_steps_data[l_idx] = normalise_one_steps_data(
                    y_standardised_first_step, bin_num
                )
            else:
                first_step = standardise_y_z_flip_gait_add_features_to_one_step(
                    first_step, global_Y_max, info, cfg
                )
                all_steps_data[l_idx] = first_step
                normalised_steps_data[l_idx] = normalise_one_steps_data(
                    first_step, bin_num
                )
            # some prep for addition of further steps
            sc_num[l_idx] = len(all_cycles[legname])
            nanvector = data_copy.loc[[SEPARATOR_IDX]]
            nanvector[:] = np.nan
            # .............................  step-loop  ................................
            for s in range(1, sc_num[l_idx], 1):
                # get step separators
                numvector = data_copy.loc[[SEPARATOR_IDX]]
                numvector[:] = s + 1
                # this_step
                this_step = data_copy.loc[
                    all_cycles[legname][s][0] : all_cycles[legname][s][1]
                ]
                if standardise_y_coordinates:
                    this_step, this_y_standardised_step = (
                        standardise_y_z_flip_gait_add_features_to_one_step(
                            this_step, global_Y_max, info, cfg
                        )
                    )
                    this_normalised_step = normalise_one_steps_data(
                        this_y_standardised_step, bin_num
                    )
                else:
                    this_step = standardise_y_z_flip_gait_add_features_to_one_step(
                        this_step, global_Y_max, info, cfg
                    )
                    this_normalised_step = normalise_one_steps_data(this_step, bin_num)
                # step separators & step-to-rest concatenation
                # => note that normalised step is already based on y-stand if required
                all_steps_data[l_idx] = add_step_separators(
                    all_steps_data[l_idx], nanvector, numvector
                )
                all_steps_data[l_idx] = pd.concat(
                    [all_steps_data[l_idx], this_step], axis=0
                )
                if standardise_y_coordinates:
                    y_standardised_steps_data[l_idx] = add_step_separators(
                        y_standardised_steps_data[l_idx], nanvector, numvector
                    )
                    y_standardised_steps_data[l_idx] = pd.concat(
                        [y_standardised_steps_data[l_idx], this_y_standardised_step],
                        axis=0,
                    )
                # this_normalised_step
                normalised_steps_data[l_idx] = add_step_separators(
                    normalised_steps_data[l_idx], nanvector, numvector
                )
                normalised_steps_data[l_idx] = pd.concat(
                    [normalised_steps_data[l_idx], this_normalised_step], axis=0
                )
        # .............................  after step-loop  ..............................
        # 1) add a column to both dfs informing about leg (important for combining legs)
        all_steps_data[l_idx] = pd.concat(
            [
                all_steps_data[l_idx],  # all_steps_data
                pd.DataFrame(
                    data=legname,
                    index=all_steps_data[l_idx].index,
                    columns=[LEG_COL],
                ),
            ],
            axis=1,
        )
        if standardise_y_coordinates:
            y_standardised_steps_data[l_idx] = pd.concat(
                [
                    y_standardised_steps_data[l_idx],  # y_standardised_steps_data
                    pd.DataFrame(
                        data=legname,
                        index=y_standardised_steps_data[l_idx].index,
                        columns=[LEG_COL],
                    ),
                ],
                axis=1,
            )
        normalised_steps_data[l_idx] = pd.concat(
            [
                normalised_steps_data[l_idx],  # normalised_steps_data
                pd.DataFrame(
                    data=legname,
                    index=normalised_steps_data[l_idx].index,
                    columns=[LEG_COL],
                ),
            ],
            axis=1,
        )
        # 2) create and save average_ and std_data
        # => note that normalised_steps_data is automatically based on y-standardisation
        #    which translates to average_data & std_data
        average_data[l_idx], std_data[l_idx] = compute_average_and_std_data(
            normalised_steps_data[l_idx], bin_num, analyse_average_y
        )
    # ................................  after leg-loop  ................................
    # 1a) create "both" sheets for all our data-formats (added to -1 idx of df_list)
    # => note that we only need only_one_valid_leg once so the first 4 functions are
    #    called with [0] to index the output-tuple and get only the df from the function
    all_steps_data = combine_legs(all_steps_data, "concatenate")[0]
    normalised_steps_data = combine_legs(normalised_steps_data, "concatenate")[0]
    average_data = combine_legs(average_data, "average")[0]
    if standardise_y_coordinates:
        y_standardised_steps_data = combine_legs(
            y_standardised_steps_data, "concatenate"
        )[0]
    std_data, only_one_valid_leg = combine_legs(std_data, "average")  # legvalidity here
    # 1b) inform user if only one leg had valid SCs (affects 3rd sheet!)
    if only_one_valid_leg:
        this_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "Only the "
            + only_one_valid_leg
            + " leg had valid step cycles after our sanity checks. "
            + "\nThe third sheet of generated Sheet (XLS) files therefore only "
            + "reflects this leg's data!"
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)
    # 2) loop to assign to results and save to xls-file sheets
    for idx, output in enumerate(OUTPUTS):
        results[output]["all_steps_data"] = all_steps_data[idx]
        results[output]["normalised_steps_data"] = normalised_steps_data[idx]
        results[output]["average_data"] = average_data[idx]
        results[output]["std_data"] = std_data[idx]
        if standardise_y_coordinates:
            results[output]["y_standardised_steps_data"] = y_standardised_steps_data[
                idx
            ]
        if idx == len(OUTPUTS) - 1:
            if not only_one_valid_leg:
                results[output]["sc_num"] = sc_num[0] + sc_num[1]
            else:
                if only_one_valid_leg == "left":
                    results[output]["sc_num"] = sc_num[0]
                elif only_one_valid_leg == "right":
                    results[output]["sc_num"] = sc_num[1]
        else:
            results[output]["sc_num"] = sc_num[idx]
        save_results_sheet(
            all_steps_data[idx],
            output,
            os.path.join(results_dir, name + ORIGINAL_XLS_FILENAME),
            only_one_valid_leg,
        )
        save_results_sheet(
            normalised_steps_data[idx],
            output,
            os.path.join(results_dir, name + NORMALISED_XLS_FILENAME),
            only_one_valid_leg,
        )
        save_results_sheet(
            average_data[idx],
            output,
            os.path.join(results_dir, name + AVERAGE_XLS_FILENAME),
            only_one_valid_leg,
        )
        save_results_sheet(
            std_data[idx],
            output,
            os.path.join(results_dir, name + STD_XLS_FILENAME),
            only_one_valid_leg,
        )
        if standardise_y_coordinates:
            save_results_sheet(
                y_standardised_steps_data[idx],
                output,
                os.path.join(results_dir, name + Y_STANDARDISED_XLS_FILENAME),
                only_one_valid_leg,
            )
    return results


# ......................................................................................
# .........  helper functions a - stand y/z, flip gait and add features  ...............
# ......................................................................................


def standardise_y_z_flip_gait_add_features_to_one_step(step, global_Y_max, info, cfg):
    """For a single step cycle's data, standardise y & z if wanted, flip y columns if wanted (to simulate equal run direction) and add features (angles & velocities)"""
    # unpack
    standardise_z_at_SC_level = cfg["standardise_z_at_SC_level"]
    standardise_z_to_a_joint = cfg["standardise_z_to_a_joint"]
    z_standardisation_joint = cfg["z_standardisation_joint"]
    flip_gait_direction = cfg["flip_gait_direction"]
    standardise_y_coordinates = cfg["standardise_y_coordinates"]
    y_standardisation_joint = cfg["y_standardisation_joint"]
    direction_joint = cfg["direction_joint"]
    # if user wanted this, normalise z (height) at step-cycle level
    step_copy = step.copy()
    if standardise_z_at_SC_level is True:
        z_cols = [col for col in step_copy.columns if col.endswith("Z")]
        if standardise_z_to_a_joint is True:
            # note the [0] here is important because it's still a list of len=1!!
            z_minimum = step_copy[z_standardisation_joint[0] + "Z"].min()
        else:
            z_minimum = min(step_copy[z_cols].min())
        step_copy[z_cols] -= z_minimum
    # no y-standardisation, just flip the step if needed, add features and return
    if standardise_y_coordinates is False:
        non_stand_step = step_copy.copy()
        if flip_gait_direction:
            non_stand_step = flip_a_steps_gait_direction(
                non_stand_step, direction_joint, global_Y_max
            )
        non_stand_step = add_features(non_stand_step, info, cfg)
        return non_stand_step

    # user wanted y-standardisation
    # => BUT: make sure that we GAIT FLIP before y-standardising!
    # => i.e. flip gait direction, y-stand and finally add features
    #    - OTHERWISE the way global_Y_max is used to flip gait direction BREAKS IT!
    else:
        non_stand_step = step_copy.copy()
        y_stand_step = step_copy.copy()
        # first flip gait direction
        if flip_gait_direction:
            non_stand_step = flip_a_steps_gait_direction(
                non_stand_step, direction_joint, global_Y_max
            )
            y_stand_step = flip_a_steps_gait_direction(
                y_stand_step, direction_joint, global_Y_max
            )
        # now y-standardise
        # => standardising to the minimum of the joint is correct because flipping
        #    ensures that all steps' y values are increasing as the step progresses
        #   - things can get weird if users are not gait flipping (and are not e.g.
        #     analysing treadmill walks where y is roughly fixed) but we informed them
        #     about this before (at top of "analyse_....-function")
        y_cols = [col for col in y_stand_step.columns if col.endswith("Y")]
        y_minimum = y_stand_step[y_standardisation_joint[0] + "Y"].min()
        y_stand_step[y_cols] -= y_minimum
        # add features & return
        non_stand_step = add_features(non_stand_step, info, cfg)
        y_stand_step = add_features(y_stand_step, info, cfg)
        return non_stand_step, y_stand_step


def flip_a_steps_gait_direction(step, direction_joint, global_Y_max):
    """Flip a step's y cols first checking if it's needed for given step with direction joint"""
    direction_joint_col_idx = step.columns.get_loc(direction_joint)
    direction_joint_mean = np.mean(step[direction_joint])
    if step.iloc[0, direction_joint_col_idx] > direction_joint_mean:  # needed?
        step = flip_y_columns(step, global_Y_max)
    return step


def flip_y_columns(step, global_Y_max):
    """Flip all y columns if walking direction was right to left
    (i.e., Y was decreasing within a step cycle)
    """
    flipped_step = step.copy()  # do everything on a copy of step
    y_cols = [col for col in step.columns if col.endswith("Y")]
    flipped_step[y_cols] = global_Y_max - flipped_step[y_cols]
    return flipped_step


def add_features(step, info, cfg):
    """Add Features, i.e. Angles & Velocities

    Note
    ----
    Since adding flexibility with user-defined feature-names (i.e., joints & angles)
    we had to consider the case of a non-bodyside-specific angle to be computed (e.g.,
    Pelvis) using bodyside-specific joints (e.g., Hip).
    => Thus the angle + Y if statements in the local functions
    => Since we loop over left, right body sides and due to how we use the
       step_copy.columns.duplicated() statement below, this means that such angles
       (and corresponding velocities & accelerations) will be computed with reference
       to the left body side by default
       -- Add this to the documentation and tell people to fix their columns if
          they want to have this differently... maybe improve in the future if wanted
       -- This leads to a slight inaccuracy/limitation in that e.g. the Pelvis Angle is
          computed w.r.t. the left side of the body always, even in the XLS sheets of
          step cycles of the right leg

    DONT GET CONFUSED ABOUT STEP'S KEY NAMES WHEN YOU LOOK AT THIS IN A YEAR
    ------------------------------------------------------------------------
    The "Y", "Z", "Angle", "Velocity" strings never start with a space, e.g.
    " Velocity" because we previously ensured that joints that are not bodyside-
    specific end with a space (and thus joint + "Velocity" is correct) and if they
    are bodyside specific legname is of LEGS_COLFORMAT, so that also ends with a
    space character
    """
    # unpack
    angles = cfg["angles"]

    step_copy = step.copy()
    for legname in LEGS_COLFORMAT:
        if angles["name"]:  # if at least 1 string in list
            step_copy = add_angles(step_copy, legname, info, cfg)
        step_copy = add_velocities(step_copy, legname, cfg)
    # I do this since it's possible that we created two, e.g., "Pelvis Angle" cols due
    # to our leg-loop # !!! NU (see in the doc above)
    step_copy = step_copy.loc[:, ~step_copy.columns.duplicated()]
    return step_copy


def add_angles(step, legname, info, cfg):
    """Feature #1: Joint Angles

    Note
    ----
    legname here is from LEGS_COLFORMAT!
    """

    # unpack
    angles = cfg["angles"]

    for a, angle in enumerate(angles["name"]):
        # initialise 2d (x/y coords) arrays
        joint_angle = np.zeros((len(step), 2), dtype=float)
        joint2 = np.zeros((len(step), 2), dtype=float)
        joint3 = np.zeros((len(step), 2), dtype=float)
        # unpack joint names for current angle
        lower_joint = angles["lower_joint"][a]
        upper_joint = angles["upper_joint"][a]
        # assign values to 2d arrays
        # ==> for each of our 3 joints, check if they are body-side-specific or not
        # angle (name)
        if angle + "Y" in step.columns:
            joint_angle[:, 0] = step[angle + "Y"]
            joint_angle[:, 1] = step[angle + "Z"]
        else:
            joint_angle[:, 0] = step[angle + legname + "Y"]
            joint_angle[:, 1] = step[angle + legname + "Z"]
        # lower joint
        if lower_joint + "Y" in step.columns:
            joint2[:, 0] = step[lower_joint + "Y"]
            joint2[:, 1] = step[lower_joint + "Z"]
        else:
            joint2[:, 0] = step[lower_joint + legname + "Y"]
            joint2[:, 1] = step[lower_joint + legname + "Z"]
        # upper joint
        if upper_joint + "Y" in step.columns:
            joint3[:, 0] = step[upper_joint + "Y"]
            joint3[:, 1] = step[upper_joint + "Z"]
        else:
            joint3[:, 0] = step[upper_joint + legname + "Y"]
            joint3[:, 1] = step[upper_joint + legname + "Z"]
        # initialise the angle vector and assign looping over timepoints
        this_angle = np.zeros(len(joint_angle))
        broken_angle_idxs = []  # initialise broken idxs-list for each angle anew
        for t in range(len(joint_angle)):
            this_angle[t], broken = compute_angle(
                joint_angle[t, :], joint2[t, :], joint3[t, :]
            )
            if broken:
                broken_angle_idxs.append(t)
        if broken_angle_idxs:
            write_angle_warning(
                step, a, angle, broken_angle_idxs, info, legname=legname
            )
        # colnames depend on bodyside-specificity
        if angle + "Y" in step.columns:
            this_colname = angle + "Angle"
        elif angle + legname + "Y" in step.columns:
            this_colname = angle + legname + "Angle"
        step[this_colname] = this_angle
    return step


def add_velocities(step, legname, cfg):
    """Feature #2: Joint Y and Angular Velocities"""

    # unpack
    joints = cfg["joints"]
    angles = cfg["angles"]
    y_acceleration = cfg["y_acceleration"]
    angular_acceleration = cfg["angular_acceleration"]
    original_legname = legname  # in case we overwrite legname to be an empty string

    # compute velocities (& acceleration if wanted) for joints
    for joint in joints:
        if joint + "Y" in step.columns:
            legname = ""
        else:
            legname = original_legname
        step[joint + legname + "Velocity"] = 0.0
        step.loc[:, joint + legname + "Velocity"] = np.gradient(
            step.loc[:, joint + legname + "Y"]
        )
        if y_acceleration:
            step[joint + legname + "Acceleration"] = 0.0
            step.loc[:, joint + legname + "Acceleration"] = np.gradient(
                step.loc[:, joint + legname + "Velocity"]
            )
    # compute velocities (& acceleration) for the angles too
    for angle in angles["name"]:
        if angle + "Angle" in step.columns:
            legname = ""
        else:
            legname = original_legname
        angle_colname = angle + legname + "Angle"
        step[angle_colname + " Velocity"] = 0.0  # spaces in colnames here!
        step.loc[:, angle_colname + " Velocity"] = np.gradient(
            step.loc[:, angle_colname]
        )
        if angular_acceleration:
            step[angle_colname + " Acceleration"] = 0.0
            step.loc[:, angle_colname + " Acceleration"] = np.gradient(
                step.loc[:, angle_colname + " Velocity"]
            )
    return step


# ......................................................................................
# .................  helper functions b - sc length normalisation  .....................
# ......................................................................................


def normalise_one_steps_data(step, bin_num):
    """Normalise all steps to be of length 25 - uses define_bins

    Important
    ---------
    The input step here is a pd dataframe that only captures ONE stepcycle!
    (concatenation happens in exportsteps function)
    """

    normalised_step = pd.DataFrame(
        data=None, index=range(bin_num), columns=step.columns
    )
    for c, col in enumerate(step.columns):
        thistrial = step[col]
        if c == 0:  # if first column, define bins anew
            bins = define_bins(int(len(thistrial)), bin_num)
        normtrial = np.zeros(bin_num)
        if isinstance(bins[0], list):  # we need to average
            for i in range(bin_num):
                normtrial[i] = np.mean(thistrial.iloc[bins[i]])
        else:  # no need to average, repeat or assign
            for i in range(bin_num):
                normtrial[i] = thistrial.iloc[bins[i]]
        normalised_step[col] = normtrial
    return normalised_step


def define_bins(triallength, bin_num):
    """Define bins to know which indices move which bin for normalisation"""
    indices = list(range(triallength))
    bins = [[] for i in range(bin_num)]
    # moving average to make the trial shorter
    if triallength > bin_num:
        idx = 0
        for i in indices:
            idx += 1
            if i % bin_num == 0:
                idx = 0
            bins[idx].append(i)  # this means that bins[0] = [0, 25, 50, etc.]
        for i in range(len(bins)):
            for j in range(len(bins[i])):
                if i == 0:
                    bins[i][j] = j
                else:  # here we make sure it gets: bins[0] = [0, 1, 2, etc.]
                    bins[i][j] = j + np.max(bins[i - 1]) + 1  # +1 bc. idx from 0
    # repeat the same values to make the trial longer
    elif triallength < bin_num:
        final_list = []
        # use remainder to fill list after triallength val until len = bin_num
        # e.g. if triallength = 17 it goes 16, 17, 0, 1, 2, 3, 4, 5, 6
        for i in range(bin_num):
            idx = indices[i % len(indices)]
            final_list.append(idx)
        final_list.sort()  # need to sort (see above for initial order)
        bins = final_list
        if (len(bins) != bin_num) | (np.max(bins) != triallength - 1):
            raise Exception("Binning bugged (shouldn't happen) - contact me.")
    # if exactly 25 points originally
    else:
        for i in range(triallength):
            bins[i] = i
    return bins


# ......................................................................................
# ....................  helper functions c - xls file stuff  ...........................
# ......................................................................................


def compute_average_and_std_data(normalised_steps_data, bin_num, analyse_average_y):
    """Export XLS tables that store all averages & std of y-coords & angles"""
    # initialise data & columns of average & std dataframes (fill vals in loop)
    initialisation_data = bin_num_to_percentages(bin_num)
    # only for 3D - make a list of the percentages list that is output
    initialisation_data = [initialisation_data]
    initialisation_columns = [SC_PERCENTAGE_COL]
    if analyse_average_y:
        cols_to_include = [
            c
            for c in normalised_steps_data.columns
            if c not in EXCLUDED_COLS_IN_AV_STD_DFS
        ]
    else:
        cols_to_include = [
            c
            for c in normalised_steps_data.columns
            if (c not in EXCLUDED_COLS_IN_AV_STD_DFS) and (not c.endswith(" Y"))
        ]
    for col in cols_to_include:
        initialisation_data.append([None] * bin_num)
        initialisation_columns.append(col)
    initialisation_data = np.asarray(initialisation_data).transpose()  # match df-shape
    # initialise dataframes
    average_data = pd.DataFrame(
        data=initialisation_data, index=range(bin_num), columns=initialisation_columns
    )
    std_data = pd.DataFrame(
        data=initialisation_data, index=range(bin_num), columns=initialisation_columns
    )
    # loop over step cycles & columns and fill df values as you go
    # => important to note that c corresponds to colidxs of original df
    #    (not of av/std dfs) - we use col strings to assign values to correct cols of
    #    our new dfs
    sc_num = len(np.where(normalised_steps_data.index == 0)[0])
    for c, col in enumerate(
        normalised_steps_data.columns
    ):  # caution! c => normalised_steps_data!!
        if col in cols_to_include:
            this_data = np.zeros([bin_num, sc_num])
            for s in range(sc_num):
                # with this_end it's bin_num & not bin_num -1 because iloc
                # does not include last index
                this_start = np.where(normalised_steps_data.index == 0)[0][s]
                this_end = np.where(normalised_steps_data.index == 0)[0][s] + bin_num
                this_data[:, s] = normalised_steps_data.iloc[this_start:this_end, c]
            average_data[col] = np.mean(this_data, axis=1)
            std_data[col] = np.std(this_data, axis=1)
    return average_data, std_data


def combine_legs(dataframe_list, combination_procedure):
    """Combine the results of left and right legs as a new df to export as Sheet 3"""
    # first check if we only have one valid leg
    # => if av/std dfs are empty they only have SC_COL
    only_one_valid_leg = False
    if combination_procedure == "concatenate":
        # list[0] is left, so right is valid & vice versa
        if dataframe_list[0].empty:
            only_one_valid_leg = "right"
        if dataframe_list[1].empty:
            only_one_valid_leg = "left"
    elif combination_procedure == "average":
        if (
            len(dataframe_list[0].columns) == 1
            and dataframe_list[0].columns[0] == SC_PERCENTAGE_COL
        ):
            only_one_valid_leg = "right"
        if (
            len(dataframe_list[1].columns) == 1
            and dataframe_list[1].columns[0] == SC_PERCENTAGE_COL
        ):
            only_one_valid_leg = "left"
    if combination_procedure == "concatenate":
        # copy dfs if only one leg, otherwise concatenate
        if only_one_valid_leg == "left":
            dataframe_list[-1] = dataframe_list[0].copy()
        elif only_one_valid_leg == "right":
            dataframe_list[-1] = dataframe_list[1].copy()
        else:
            infovector = dataframe_list[0].iloc[0, :]
            infovector = pd.DataFrame(infovector).T
            infovector.index = [SEPARATOR_IDX]
            infovector[:] = "leg-change"
            nanvector = dataframe_list[0].iloc[0, :]
            nanvector = pd.DataFrame(nanvector).T
            nanvector.index = [SEPARATOR_IDX]
            nanvector[:] = np.nan
            # concatenation (similar separators as between stepcycles)
            # ignoring warnings here temporarily
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                dataframe_list[-1] = pd.concat(
                    [dataframe_list[0], nanvector], axis=0
                )  # list0!
            dataframe_list[-1] = pd.concat(
                [dataframe_list[-1], infovector], axis=0
            )  # -1 !
            dataframe_list[-1] = pd.concat([dataframe_list[-1], nanvector], axis=0)
            dataframe_list[-1] = pd.concat(
                [dataframe_list[-1], dataframe_list[1]], axis=0
            )
            # reorder columns
            for i in range(len(dataframe_list)):
                cols = REORDER_COLS_IN_STEP_NORMDATA
                dataframe_list[i] = dataframe_list[i][
                    cols + [c for c in dataframe_list[i].columns if c not in cols]
                ]
    elif combination_procedure == "average":
        # we can assign as we do here since dataframe_lists' dfs are average or std dfs
        # => note that we purposely check against SC col above for these to make sure
        if only_one_valid_leg == "left":
            dataframe_list[-1] = dataframe_list[0].copy()
        elif only_one_valid_leg == "right":
            dataframe_list[-1] = dataframe_list[1].copy()
        else:
            # initialise both df with same idx and cols as left df
            dataframe_list[-1] = pd.DataFrame(
                data=None,
                index=dataframe_list[0].index,
                columns=dataframe_list[0].columns,
            )
            for col in dataframe_list[0].columns:
                this_data = np.zeros([len(dataframe_list[0].index), 2])
                this_data[:, 0] = np.asarray(dataframe_list[0][col])
                this_data[:, 1] = np.asarray(dataframe_list[1][col])
                dataframe_list[-1][col] = np.mean(this_data, axis=1)
    return dataframe_list, only_one_valid_leg


# ......................................................................................
# ....................  helper functions d - miscellaneous  ............................
# ......................................................................................


def flatten_all_cycles(all_cycles):
    """Extract all runs' SC latencies and create a list of len=SC_num for each leg"""
    flattened_cycles = {"left": [], "right": []}
    for legname in LEGS:
        if all_cycles[legname]:  # can be None
            for run_cycles in all_cycles[legname]:
                for cycle in run_cycles:
                    flattened_cycles[legname].append(cycle)
    return flattened_cycles


def delete_previous_xlsfiles(name, results_dir):
    """
    Check and delete previously stored excel files (important since appending sheets)
    """
    all_filenames = [
        ORIGINAL_XLS_FILENAME,
        NORMALISED_XLS_FILENAME,
        AVERAGE_XLS_FILENAME,
        STD_XLS_FILENAME,
    ]
    for filename in all_filenames:
        fullfilepath = results_dir + name + filename + ".xlsx"
        if os.path.exists(fullfilepath):
            os.remove(fullfilepath)


def save_results_sheet(dataframe, sheet, fullfilepath, only_one_valid_leg):
    """Save a xls sheet of given df"""
    fullfilepath = fullfilepath + ".xlsx"
    if os.path.exists(fullfilepath):
        with pd.ExcelWriter(fullfilepath, mode="a") as writer:
            if (only_one_valid_leg == "left" and sheet == "right") or (
                only_one_valid_leg == "right" and sheet == "left"
            ):
                pd.DataFrame(data=None).to_excel(writer, sheet_name=sheet, index=False)
            else:
                dataframe.to_excel(writer, sheet_name=sheet, index=False)
    else:
        with pd.ExcelWriter(fullfilepath) as writer:
            if (only_one_valid_leg == "left" and sheet == "right") or (
                only_one_valid_leg == "right" and sheet == "left"
            ):
                pd.DataFrame(data=None).to_excel(writer, sheet_name=sheet, index=False)
            else:
                dataframe.to_excel(writer, sheet_name=sheet, index=False)


def add_step_separators(dataframe, nanvector, numvector):
    """Add nan & num vector separators between step cycles to dataframes"""
    dataframe = pd.concat([dataframe, nanvector], axis=0)  # nan
    dataframe = pd.concat([dataframe, numvector], axis=0)  # num
    dataframe = pd.concat([dataframe, nanvector], axis=0)  # nan
    return dataframe



import pytest
from autogaita.common2D.common2D_utils import extract_info, find_number
import os
import pytest


# %%..............................  fixtures  ..........................................
# => CALL FIXTURES _fixture SINCE WE HAVE A FUNCTION CALLED extract_info!
@pytest.fixture
def fixture_extract_info(tmp_path):
    info = {}
    info["mouse_num"] = 15
    info["run_num"] = 3
    info["name"] = "ID " + str(info["mouse_num"]) + " - Run " + str(info["run_num"])
    info["results_dir"] = os.path.join(tmp_path, info["name"])
    return info


@pytest.fixture
def fixture_extract_folderinfo():
    folderinfo = {}
    folderinfo["root_dir"] = "tests/test_data/dlc_data"
    folderinfo["sctable_filename"] = (
        "correct_annotation_table.xlsx"  # has to be an excel file
    )
    folderinfo["data_string"] = "SIMINewOct"
    folderinfo["beam_string"] = "BeamTraining"
    folderinfo["premouse_string"] = "Mouse"
    folderinfo["postmouse_string"] = "_25mm"
    folderinfo["prerun_string"] = "run"
    folderinfo["postrun_string"] = "-6DLC"
    return folderinfo


@pytest.fixture
def fixture_extract_cfg():
    cfg = {}
    cfg["sampling_rate"] = 100
    cfg["subtract_beam"] = True
    cfg["dont_show_plots"] = True
    cfg["convert_to_mm"] = True
    cfg["pixel_to_mm_ratio"] = 3.76
    cfg["x_sc_broken_threshold"] = 200
    cfg["y_sc_broken_threshold"] = 50
    cfg["x_acceleration"] = True
    cfg["angular_acceleration"] = True
    cfg["save_to_xls"] = True
    cfg["bin_num"] = 25
    cfg["plot_SE"] = True
    cfg["standardise_y_at_SC_level"] = False
    cfg["standardise_y_to_a_joint"] = True
    cfg["y_standardisation_joint"] = ["Knee"]
    cfg["plot_joint_number"] = 3
    cfg["color_palette"] = "viridis"
    cfg["legend_outside"] = True
    cfg["invert_y_axis"] = True
    cfg["flip_gait_direction"] = True
    cfg["analyse_average_x"] = True
    cfg["standardise_x_coordinates"] = True
    cfg["x_standardisation_joint"] = ["Hind paw tao"]
    cfg["coordinate_standardisation_xls"] = ""
    cfg["hind_joints"] = ["Hind paw tao", "Ankle", "Knee", "Hip", "Iliac Crest"]
    cfg["fore_joints"] = [
        "Front paw tao ",
        "Wrist ",
        "Elbow ",
        "Lower Shoulder ",
        "Upper Shoulder ",
    ]
    cfg["beam_col_left"] = ["BeamLeft"]  # BEAM_COL_LEFT & _RIGHT must be lists of len=1
    cfg["beam_col_right"] = ["BeamRight"]
    cfg["beam_hind_jointadd"] = ["Tail base ", "Tail center ", "Tail tip "]
    cfg["beam_fore_jointadd"] = ["Nose ", "Ear base "]
    cfg["angles"] = {
        "name": ["Ankle ", "Knee ", "Hip "],
        "lower_joint": ["Hind paw tao ", "Ankle ", "Knee "],
        "upper_joint": ["Knee ", "Hip ", "Iliac Crest "],
    }
    return cfg


# %%..............................  tests  ..........................................
# 1) Test extract_info function
def test_extract_info_function(fixture_extract_folderinfo):
    # => IT IS CORRECT THAT EXTRACT_INFO RETURNS lists of integers!
    # a. smoke test standard case
    info = extract_info("DLC", fixture_extract_folderinfo)
    assert info["mouse_num"] == [15]
    assert info["run_num"] == [3]
    # b. test that underscores/hyphens are added
    folderinfo = fixture_extract_folderinfo
    folderinfo["postmouse_string"] = "25mm"
    folderinfo["postrun_string"] = "6DLC"
    info = extract_info("DLC", folderinfo)
    assert info["mouse_num"] == [15]
    assert info["run_num"] == [3]
    # c. test that leading zeros are removed
    folderinfo["root_dir"] = "tests/test_data/dlc_data/test_data/leading_zeros/"
    info = extract_info("DLC", folderinfo)
    assert info["mouse_num"] == [12, 12]
    # order of this might break in some test env
    assert info["run_num"] == [4, 3]
    assert info["leading_mouse_num_zeros"] == [False, False]
    assert info["leading_run_num_zeros"] == ["00", False]


# 2) Test find number function - also for leading zeros
def test_find_number_valid_case():
    fullstring = "Mouse1_Run002"
    prestring = "Mouse"
    poststring = "_Run"
    result = find_number(fullstring, prestring, poststring)
    assert result == (1, False)


def test_find_number_prestring_not_found():
    fullstring = "Mouse001_Run002"
    prestring = "Cat"
    poststring = "_Run"
    with pytest.raises(ValueError):
        find_number(fullstring, prestring, poststring)


def test_find_number_poststring_not_found():
    fullstring = "Mouse001_Run002"
    prestring = "Mouse"
    poststring = "_Walk"
    with pytest.raises(ValueError):
        find_number(fullstring, prestring, poststring)


def test_find_number_leading_zeros():
    fullstring = "Mouse001Run002"
    prestring = "Mouse"
    poststring = "Run"
    result = find_number(fullstring, prestring, poststring)
    assert result == (1, "00")



from autogaita.universal3D.universal3D_3_analysis import (
    standardise_y_z_flip_gait_add_features_to_one_step,
    add_features,
)
from hypothesis import HealthCheck, given, settings, strategies as st
import pytest
import pandas as pd
import pandas.testing as pdt
import numpy as np
import os


# %%................................  fixtures  ........................................
@pytest.fixture
def extract_info(tmp_path):
    info = {}
    info["name"] = "SK"
    info["results_dir"] = os.path.join(tmp_path, info["name"])
    return info


@pytest.fixture
def extract_folderinfo():
    folderinfo = {}
    folderinfo["root_dir"] = "tests/test_data/universal3D_data/test_data"
    folderinfo["sctable_filename"] = "SC Latency Table.xlsx"
    folderinfo["postname_string"] = ""
    return folderinfo


@pytest.fixture
def extract_cfg():
    # note space in end of "Midfoot, left " must be here bc. we don't run our fix and
    # check cfg function of 1_prep_
    cfg = {}
    cfg["sampling_rate"] = 100  # base cfg
    cfg["dont_show_plots"] = True
    cfg["y_acceleration"] = True
    cfg["angular_acceleration"] = True
    cfg["bin_num"] = 25
    cfg["plot_SE"] = True
    cfg["standardise_z_at_SC_level"] = True
    cfg["standardise_z_to_a_joint"] = True
    cfg["z_standardisation_joint"] = ["Midfoot, left "]
    cfg["plot_joint_number"] = 5
    cfg["color_palette"] = "Set2"
    cfg["legend_outside"] = True
    cfg["flip_gait_direction"] = True
    cfg["analyse_average_y"] = True
    cfg["standardise_y_coordinates"] = True
    cfg["y_standardisation_joint"] = ["Midfoot, left "]
    cfg["coordinate_standardisation_xls"] = ""
    cfg["joints"] = [
        "Midfoot",
        "Ankle",
        "Knee",
        "Hip",
    ]
    cfg["angles"] = {
        "name": ["Ankle", "Knee"],
        "lower_joint": ["Midfoot", "Ankle"],
        "upper_joint": ["Knee", "Hip"],
    }
    cfg["direction_joint"] = "Midfoot, left Y"
    return cfg


sample_step_len = 10  # 10 datapoints (e.g. time)
sample_step_col_num = 20  # 10 for each Y and Z


@pytest.fixture
def sample_step():
    joint_strings = [
        "Midfoot, left",
        "Ankle, left",
        "Knee, left",
        "Hip, left",
        "Midfoot, right",
        "Ankle, right",
        "Knee, right",
        "Hip, right",
        "Shoulder",
        "Neck",
    ]
    sample_step = {}
    for joint in joint_strings:
        for coord in [" Y", " Z"]:
            sample_step[joint + coord] = list(
                np.random.randint(1, 101, sample_step_len)
            )
    sample_step["Time"] = np.arange(sample_step_len)
    return pd.DataFrame(sample_step)


# ............................  for property tests  ....................................
# 1. define the sample_steps_data strategy using random ints as data
sample_steps_data_strategy = st.lists(
    st.lists(
        st.integers(min_value=-1000, max_value=1000),
        min_size=sample_step_col_num,  # 20 cols (10 for each Y and Z)
        max_size=sample_step_col_num,
    ),
    min_size=sample_step_len,  # 10 rows
    max_size=sample_step_len,
)


# 2. create a custom decorator that can be used instead of writing given/settings
#    always
def sample_data_for_property_tests(func):
    return settings(suppress_health_check=(HealthCheck.function_scoped_fixture,))(
        given(sample_steps_data=sample_steps_data_strategy)(func)
    )


# ..................................  tests  .........................................
# %% workflow step #3 - y-flipping, y-stand, features, df-creation & exports


def test_standardise_z_at_SC_level(sample_step, extract_info, extract_cfg):
    extract_cfg["standardise_z_at_SC_level"] = True
    extract_cfg["standardise_z_to_a_joint"] = False
    extract_cfg["flip_gait_direction"] = False
    extract_cfg["standardise_y_coordinates"] = False
    z_cols = [col for col in sample_step.columns if col.endswith("Z")]
    function_step = standardise_y_z_flip_gait_add_features_to_one_step(
        sample_step, 10, extract_info, extract_cfg
    )
    sample_step = add_features(
        sample_step, extract_info, extract_cfg
    )  # otherwise df-shape mismatch
    steps_global_z_minimum = sample_step[z_cols].min().min()  # global == all joints
    expected_step = sample_step.copy()
    expected_step[z_cols] -= steps_global_z_minimum
    pdt.assert_frame_equal(function_step, expected_step)


# because we have random integers in step, we might get a near-zero vectors when
# computing angles - ignore those warnings
@pytest.mark.filterwarnings("ignore:invalid value")
def test_flip_gait_direction(sample_step, extract_info, extract_cfg):
    # prepare some vars
    extract_cfg["flip_gait_direction"] = True
    extract_cfg["standardise_y_coordinates"] = False
    global_Y_max = 10
    to_be_flipped_step = sample_step.copy()
    to_not_be_flipped_step = sample_step.copy()
    y_cols = [col for col in sample_step.columns if col.endswith("Y")]
    # step sample is random integers, so first fix the direction of y-values, either
    # increasing or decreasing (decreasing must be flipped)
    for col in y_cols:
        to_be_flipped_step[col] = to_be_flipped_step[col].sort_values(
            ascending=False, ignore_index=True
        )
        to_not_be_flipped_step[col] = to_not_be_flipped_step[col].sort_values(
            ascending=True, ignore_index=True
        )
    to_be_flipped_step = standardise_y_z_flip_gait_add_features_to_one_step(
        to_be_flipped_step, global_Y_max, extract_info, extract_cfg
    )
    to_not_be_flipped_step = standardise_y_z_flip_gait_add_features_to_one_step(
        to_not_be_flipped_step, global_Y_max, extract_info, extract_cfg
    )
    # first test if the y-values are flipped (increasing y-cols progressively)
    for col in y_cols:
        assert to_be_flipped_step[col][0] < to_be_flipped_step[col].mean()
    # now, test that if you reverse the impact of global_y_max you get your original df
    # 1. reverse subtraction of global y max before comparison for df-equivalence
    # => e.g. 10 - 2 = 8 || reverse via: 10 - 8 = 2!
    to_be_flipped_step[y_cols] = global_Y_max - to_be_flipped_step[y_cols]
    # 2. reverting of 1. changes y-values to now be decreasing (2 4 8 becomes 8 4 2)
    # => sooo manually make it so that it is increasing again and then you should have
    #    equal dfs
    for col in y_cols:
        to_be_flipped_step[col] = to_be_flipped_step[col].sort_values(
            ascending=True, ignore_index=True
        )
    # 3. remove the cols that were created by add_features, because those won't match
    #    (understandably)
    # fmt:off
    cols_to_drop = [col for col in to_be_flipped_step.columns if not (col.endswith("Z") or col.endswith("Y"))]
    for df in [to_be_flipped_step, to_not_be_flipped_step]:
        df.drop(columns=cols_to_drop, inplace=True)
    # fmt:on
    pdt.assert_frame_equal(to_be_flipped_step, to_not_be_flipped_step)


# because we have random integers in step, we might get a near-zero vectors when
# computing angles - ignore those warnings
@pytest.mark.filterwarnings("ignore:invalid value")
@sample_data_for_property_tests
def test_standardise_y_coordinates_no_gait_flipping(
    sample_step, extract_info, extract_cfg, sample_steps_data
):
    extract_cfg["standardise_y_coordinates"] = True
    extract_cfg["flip_gait_direction"] = False
    y_cols = [col for col in sample_step.columns if col.endswith("Y")]
    # prep data
    # => because we are property testing, insert the hypothesis-generated data into the
    #    sample_step df (which has the correct columns)
    # => the -1 in pd.df line is because we first want to exclude the time column...
    sample_step = pd.DataFrame(columns=sample_step.columns[:-1], data=sample_steps_data)
    sample_step["Time"] = np.arange(len(sample_step))  # ... to include it manually
    non_stand_step, y_stand_step = standardise_y_z_flip_gait_add_features_to_one_step(
        sample_step, 10, extract_info, extract_cfg
    )
    steps_y_min = (
        sample_step[extract_cfg["y_standardisation_joint"][0] + "Y"].min().min()
    )
    non_stand_step[y_cols] -= steps_y_min
    pdt.assert_frame_equal(non_stand_step, y_stand_step)


def test_standardise_y_coordinates_gait_flipping(
    sample_step, extract_info, extract_cfg
):
    # prep vars
    extract_cfg["standardise_y_coordinates"] = True
    extract_cfg["flip_gait_direction"] = True
    global_Y_max = 10
    y_cols = [col for col in sample_step.columns if col.endswith("Y")]

    # ---------------
    # NOTE FOR MYSELF
    # => look at this again - returned math domain errors if property testing
    # => uncomment the sample_step =... line & use the @sample_data.. fixture as in the
    #    test above to have another look
    # => because we are property testing, insert the hypothesis-generated data into the
    #    sample_step df (which has the correct columns)
    # sample_step = pd.DataFrame(columns=sample_step.columns, data=sample_steps_data)
    # ---------------

    # run function on a to-be-flipped step
    to_be_flipped_step = sample_step.copy()
    for col in y_cols:
        to_be_flipped_step[col] = to_be_flipped_step[col].sort_values(
            ascending=False, ignore_index=True
        )
    non_stand_step, y_stand_step = standardise_y_z_flip_gait_add_features_to_one_step(
        to_be_flipped_step, global_Y_max, extract_info, extract_cfg
    )
    steps_y_min = (
        non_stand_step[extract_cfg["y_standardisation_joint"][0] + "Y"].min().min()
    )
    reverted_step = y_stand_step.copy()
    reverted_step[y_cols] += steps_y_min
    pdt.assert_frame_equal(reverted_step, non_stand_step)



# Configuration file for DLC/SLEAP-related global constants

# 1 - preparation
DIRECTION_DLC_THRESHOLD = 0.95  # (dlc) confidence used for direction-detection
FILE_ID_STRING_ADDITIONS = ["", "-", "_"]  # (dlc) postrun/postnum string additions

# 2 - sc extraction
SCXLS_MOUSECOLS = [
    "Mouse",
    "mouse",
    "Fly",
    "fly",
    "Animal",
    "animal",
    "Subject",
    "subject",
    "ID",
    "id",
]  # SC XLS info
SCXLS_RUNCOLS = ["Run", "run", "Runs", "runs", "Trial", "trial", "Trials", "trials"]
SCXLS_SCCOLS = ["SC Number", "SC number", "sc number", "SC Num", "sc num", "SC num"]
SWINGSTART_COL = "Swing (ti)"
STANCEEND_COL = "Stance (te)"

# 3 - analysis
ORIGINAL_XLS_FILENAME = " - Original Stepcycles"  # filenames of sheet exports
NORMALISED_XLS_FILENAME = " - Normalised Stepcycles"
AVERAGE_XLS_FILENAME = " - Average Stepcycle"
STD_XLS_FILENAME = " - Standard Devs. Stepcycle"
X_STANDARDISED_XLS_FILENAME = " - X-Standardised Stepcycles"

# 4 - plots
SC_LAT_LEGEND_FONTSIZE = 8



# %% imports
from autogaita.common2D.common2D_constants import FILE_ID_STRING_ADDITIONS
from autogaita.resources.utils import try_to_run_gaita, write_issues_to_textfile
import os
import copy
import numpy as np
import tkinter as tk

# %% constants
from autogaita.resources.constants import TIME_COL


def run_singlerun_in_multirun(tracking_software, idx, info, folderinfo, cfg):
    """When performing a multirun, either via Batch Analysis in GUI or batchrun scripts, run the analysis for a given dataset"""
    # extract and pass info of this mouse/run or name (also update resdir)
    this_info = {}
    keynames = info.keys()
    for keyname in keynames:
        # if the key is leading mouse/run num zeros, only pass it if it is not False
        # => we want these to not be keys in this_info if it is False due to how
        #    leading zeros are handled in the prep script
        if "leading_" in keyname:
            if info[keyname][idx] is not False:
                this_info[keyname] = info[keyname][idx]
        else:  # pass as is for all other keys
            this_info[keyname] = info[keyname][idx]
        if cfg["results_dir"]:
            this_info["results_dir"] = os.path.join(
                cfg["results_dir"], this_info["name"]
            )
        else:
            this_info["results_dir"] = os.path.join(
                folderinfo["root_dir"], "Results", this_info["name"]
            )
    # make a deep copy of cfg that used in each run, otherwise changes to the cfg dict
    # would translate to subsequent runs
    # ==> see https://stackoverflow.com/questions/2465921/
    #         how-to-copy-a-dictionary-and-only-edit-the-copy
    this_cfg = copy.deepcopy(cfg)
    # important to only pass this_info to main script here (1 run at a time!)
    try_to_run_gaita(tracking_software, this_info, folderinfo, this_cfg, True)


def extract_info(tracking_software, folderinfo, in_GUI=False):
    """Prepare a dict of lists that include unique infos for each dataset in a folder"""

    # unpack
    root_dir = folderinfo["root_dir"]
    premouse_string = folderinfo["premouse_string"]
    postmouse_string = folderinfo["postmouse_string"]
    prerun_string = folderinfo["prerun_string"]
    postrun_string = folderinfo["postrun_string"]
    # IMPORTANT
    # ---------
    # different file types based on which software did the tracking!
    if tracking_software == "DLC":
        file_type = ".csv"
    elif tracking_software == "SLEAP":
        file_type = ".h5"
    # prepare output info dict and run
    info = {
        "name": [],
        "mouse_num": [],
        "run_num": [],
        "leading_mouse_num_zeros": [],
        "leading_run_num_zeros": [],
    }
    for filename in os.listdir(folderinfo["root_dir"]):
        # make sure we don't get wrong files
        if (
            (premouse_string in filename)
            & (prerun_string in filename)
            & (filename.endswith(file_type))
        ):
            # ID number - fill in "_" for user if needed
            this_mouse_num = False
            for string_addition in FILE_ID_STRING_ADDITIONS:
                try:
                    candidate_postmouse_string = string_addition + postmouse_string
                    this_mouse_num, leading_mouse_num_zeros = find_number(
                        filename,
                        premouse_string,
                        candidate_postmouse_string,
                    )
                except:
                    pass
            if this_mouse_num is False:
                no_ID_num_found_msg = (
                    "Unable to extract ID numbers from file identifiers! "
                    + "Check unique subject [B] and unique task [C] identifiers"
                )
                # if we're in the GUI, show an error message & stop here
                if in_GUI:
                    tk.messagebox.showerror(
                        title="No ID number found!", message=no_ID_num_found_msg
                    )
                    return
            # Do the same for run number
            this_run_num = False
            for string_addition in FILE_ID_STRING_ADDITIONS:
                try:
                    candidate_postrun_string = string_addition + postrun_string
                    this_run_num, leading_run_num_zeros = find_number(
                        filename, prerun_string, candidate_postrun_string
                    )
                except:
                    pass
            if this_run_num is False:
                no_run_num_found_msg = (
                    "Unable to extract trial numbers from file identifiers! "
                    + "Check unique trial [D] and unique camera [E] identifiers"
                )
                if in_GUI:
                    tk.messagebox.showerror(
                        title="No Trial number found!", message=no_run_num_found_msg
                    )
                    return
            # if we found both an ID and a run number, create this_name & add to dict
            if this_mouse_num and this_run_num:  # are truthy since not False if found
                this_name = "ID " + str(this_mouse_num) + " - Run " + str(this_run_num)
                if this_name not in info["name"]:  # no data/beam duplicates here
                    info["name"].append(this_name)
                    info["mouse_num"].append(this_mouse_num)
                    info["run_num"].append(this_run_num)
                    # note that leading_zeros are always appended but only transfered
                    # to this_info dicts in singlerun_from_multi if not False
                    info["leading_mouse_num_zeros"].append(leading_mouse_num_zeros)
                    info["leading_run_num_zeros"].append(leading_run_num_zeros)
    # this might happen if user entered wrong identifiers or folder
    if len(info["name"]) < 1:
        no_files_message = (
            f"Unable to find any files at {root_dir}!" + "\ncheck your inputs!"
        )
        if in_GUI:
            tk.messagebox.showerror(
                title="No files found!",
                message=no_files_message,
            )
        else:
            print(no_files_message)
    return info

    # SLEAP BACKUP
    # => Note this previously was taking tracking_software input back when SLEAP was
    #    not identical to DLC - keep this here in case you need it at some point (note
    #    there is another return info below this)
    # elif tracking_software == "SLEAP":
    #     # unpack
    #     root_dir = folderinfo["root_dir"]
    #     data_string = folderinfo["data_string"]
    #     # prepare output info dict and run
    #     info = {"name": []}
    #     for filename in os.listdir(root_dir):
    #         if data_string + ".csv" in filename:
    #             info["name"].append(filename.split(data_string + ".csv")[0])
    #         if data_string + ".h5" in filename:
    #             info["name"].append(filename.split(data_string + ".h5")[0])
    #     info["name"] = list(set(info["name"]))  # no duplicates
    # return info


def find_number(fullstring, prestring, poststring):
    """Find (mouse/run) number based on user-defined strings in filenames"""
    start_idx = fullstring.find(prestring) + len(prestring)
    end_idx = fullstring.find(poststring)
    # handle leading zeros
    leading_zeros = ""
    while fullstring[start_idx] == "0" and start_idx < end_idx - 1:
        start_idx += 1
        leading_zeros += "0"
    if leading_zeros == "":
        leading_zeros = False  # important that it is False if none present
    return int(fullstring[start_idx:end_idx]), leading_zeros


# ...........................  SC extraction helpers  ..................................
def check_cycle_out_of_bounds(all_cycles):
    """Check if user provided SC latencies that were not in video/data bounds"""
    clean_cycles = None
    for cycle in all_cycles:
        # below checks if values are any type of int (just in case this should
        # for some super random reason change...)
        if isinstance(cycle[0], (int, np.integer)) & isinstance(
            cycle[1], (int, np.integer)
        ):
            if clean_cycles is None:
                clean_cycles = []
            clean_cycles.append(cycle)
    return clean_cycles


def check_cycle_duplicates(all_cycles):
    """Check if there are any duplicate SC latencies.
    This would break our plotting functions, which use .loc on all_steps_data - thus,
    all indices of all_cycles have to be unique. If any duplicates found, add one
    datapoint to the start latency.
    """
    for c, cycle in enumerate(all_cycles):
        if c > 0:
            if cycle[0] == all_cycles[c - 1][1]:
                all_cycles[c][0] += 1
    return all_cycles


def check_cycle_order(all_cycles, info):
    """Check if user input flawed SC latencies

    Two cases
    1. Start latency earlier than end latency of previous SC
    2. End latency earlier then start latency of current SC
    """
    clean_cycles = []
    current_max_time = 0
    for c, cycle in enumerate(all_cycles):
        if cycle[0] > current_max_time:
            if cycle[1] > cycle[0]:
                clean_cycles.append(cycle)  # only append if both tests passed
                current_max_time = cycle[1]
            else:
                this_message = (
                    "\n***********\n! WARNING !\n***********\n"
                    + "SC #"
                    + str(c + 1)
                    + " has a later start than end latency - Skipping!"
                )
                print(this_message)
                write_issues_to_textfile(this_message, info)
        else:
            this_message = (
                "\n***********\n! WARNING !\n***********\n"
                + "SC #"
                + str(c + 1)
                + " has an earlier start than previous SC's end latency - Skipping!"
            )
            print(this_message)
            write_issues_to_textfile(this_message, info)
    return clean_cycles


def check_differing_angle_joint_coords(all_cycles, data, info, cfg):
    """Check if none of the joints used for angle computations later have equal values (since this would lead to math.domain errors due to floating point precision)"""

    # Note
    # ----
    # In theory, I could fix this programatically in the add_angle function, but I feel
    # like joint-coords should not often be exactly equal like this in a meaningful way
    # We can still change it in the future.

    # unpack
    angles = cfg["angles"]

    clean_cycles = None
    for c, cycle in enumerate(all_cycles):  # for each SC
        cycle = check_a_single_cycle_for_joint_coords(cycle, angles, data, c, info)
        if cycle:  # if cycle was not valid (equal-joint-coords) this returns None
            if clean_cycles == None:
                clean_cycles = [cycle]  # also makes a 2xscs list of lists
            else:
                clean_cycles.append(cycle)
    return clean_cycles


def check_a_single_cycle_for_joint_coords(cycle, angles, data, c, info):
    for a in range(len(angles["name"])):  # for each angle configuration
        # prepare a dict that has only the data of this angle config's joints
        this_angle_data = {"name": [], "lower_joint": [], "upper_joint": []}
        for key in this_angle_data.keys():
            this_joint = angles[key][a]
            this_angle_data[key] = np.array(
                [data[this_joint + "x"], data[this_joint + "y"]]
            )
        # now check if any of the joints have the same coord at any idx
        for idx in range(cycle[0], cycle[1]):
            if (
                np.array_equal(
                    this_angle_data["name"][:, idx],
                    this_angle_data["lower_joint"][:, idx],
                )
                or np.array_equal(
                    this_angle_data["name"][:, idx],
                    this_angle_data["upper_joint"][:, idx],
                )
                or np.array_equal(
                    this_angle_data["lower_joint"][:, idx],
                    this_angle_data["upper_joint"][:, idx],
                )
            ):
                this_message = (
                    "\n***********\n! WARNING !\n***********\n"
                    + f"SC #{c + 1} has equal joint coordinates at "
                    + f"{round(data[TIME_COL][idx],4)}s:"
                    + "\n\nAngle - [x y]:\n"
                    + angles["name"][a]
                    + " - "
                    + str(this_angle_data["name"][:, idx])
                    + "\nLower joint: "
                    + angles["lower_joint"][a]
                    + " - "
                    + str(this_angle_data["lower_joint"][:, idx])
                    + "\nUpper joint: "
                    + angles["upper_joint"][a]
                    + " - "
                    + str(this_angle_data["upper_joint"][:, idx])
                    + "\nRemoving the SC from "
                    + f"{round(data[TIME_COL][cycle[0]], 4)}-"
                    + f"{round(data[TIME_COL][cycle[1]], 4)}s"
                )
                print(this_message)
                write_issues_to_textfile(this_message, info)
                return None  # removes this SC
    return cycle  # if we never returned None, this SC is valid


def check_tracking_xy_thresholds(all_cycles, data, info, cfg):
    """Check if any x/y column of any joint has broken datapoints"""
    # unpack
    convert_to_mm = cfg["convert_to_mm"]
    x_sc_broken_threshold = cfg["x_sc_broken_threshold"]
    y_sc_broken_threshold = cfg["y_sc_broken_threshold"]
    pixel_to_mm_ratio = cfg["pixel_to_mm_ratio"]
    hind_joints = cfg["hind_joints"]
    if convert_to_mm:
        x_sc_broken_threshold = x_sc_broken_threshold / pixel_to_mm_ratio
        y_sc_broken_threshold = y_sc_broken_threshold / pixel_to_mm_ratio
    columns = []
    clean_cycles = None
    for joint in hind_joints:
        columns.append(joint + "x")
        columns.append(joint + "y")
    for c, cycle in enumerate(all_cycles):
        exclude_this_cycle = False  # reset
        for col in columns:
            if col.endswith("x"):
                this_threshold = x_sc_broken_threshold
            elif col.endswith("y"):
                this_threshold = y_sc_broken_threshold
            this_data = data.loc[cycle[0] : cycle[1], col]
            for i in range(len(this_data) - 1):
                if (this_data.iloc[i + 1] > (this_data.iloc[i] + this_threshold)) | (
                    this_data.iloc[i + 1] < (this_data.iloc[i] - this_threshold)
                ):
                    exclude_this_cycle = True
        if exclude_this_cycle == True:
            this_message = (
                "\n...excluding SC #"
                + str(c + 1)
                + " - Tracking failed (coordinate-jump larger than x/y threshold)!"
            )
            print(this_message)
            write_issues_to_textfile(this_message, info)
        else:
            if clean_cycles == None:
                clean_cycles = [cycle]  # also makes a 2xscs list of lists
            else:
                clean_cycles.append(cycle)
    return clean_cycles


def check_tracking_SLEAP_nans(all_cycles, data, info, cfg):
    """In SLEAP if tracking fails it generates NaNs - make sure we don't have those in any SC in any joint or angle-joint"""
    # unpack
    hind_joints = cfg["hind_joints"]
    angles = cfg["angles"]
    # all joints to test for NaNs
    all_joints = hind_joints
    for key in angles.keys():
        all_joints += angles[key]
    # columns to check
    columns = []
    for joint in all_joints:
        columns.append(joint + "x")
        columns.append(joint + "y")
    # check for NaNs
    clean_cycles = None
    for c, cycle in enumerate(all_cycles):
        exclude_this_cycle = False  # reset
        for col in columns:
            this_data = data.loc[cycle[0] : cycle[1], col]
            if any(this_data.isna()):
                exclude_this_cycle = True
                NaN_joint = col
                break
        if exclude_this_cycle == True:
            this_message = (
                "\n...excluding SC #"
                + str(c + 1)
                + f" - Tracking failed (NaN found at {NaN_joint})!"
            )
            print(this_message)
            write_issues_to_textfile(this_message, info)
        else:
            if clean_cycles == None:
                clean_cycles = [cycle]  # also makes a 2xscs list of lists
            else:
                clean_cycles.append(cycle)
    return clean_cycles


def handle_issues(condition, info):
    """Handle different kind of issues with step-cycles (& the table)"""
    # 1: can also occur bc. all scs when dlc failed
    if condition == "scs_invalid":
        this_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "Skipped since all SCs invalid!"
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)
    # 2: no SCs were provided in XLS table
    elif condition == "no_scs":
        this_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "Skipped since no SCs in Annotation Table!"
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)
    # 3: the mouse was not included in XLS table
    elif condition == "no_mouse":
        this_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "Skipped since ID not in Annotation Table!"
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)
    # 4: user entered wrong column names in XLS table
    elif condition == "wrong_scxls_colnames":
        this_message = (
            "\n******************\n! CRITICAL ERROR !"
            + "\n******************\n"
            + "Annotation Table's Column Names are wrong!\n"
            + "Check Instructions!"
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)
    # 5:
    elif condition == "double_mouse":
        this_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "Skipped since ID found more than once in "
            + "Annotation Table!"
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)
    return



import platform
from importlib import resources
from PIL import Image
from customtkinter import CTkImage


# ...............................  general gui stuff  ..................................
def configure_the_icon(root):
    """Configure the icon - in macos it changes the dock icon, in windows it changes
    all windows titlebar icons (taskbar cannot be changed without converting to exe)
    """
    if platform.system().startswith("Darwin"):
        try:
            from Cocoa import NSApplication, NSImage
        except ImportError:
            print("Unable to import pyobjc modules")
        else:
            with resources.path("autogaita.resources", "icon.icns") as icon_path:
                ns_application = NSApplication.sharedApplication()
                logo_ns_image = NSImage.alloc().initWithContentsOfFile_(str(icon_path))
                ns_application.setApplicationIconImage_(logo_ns_image)
    elif platform.system().startswith("win"):
        with resources.path("autogaita.resources", "icon.ico") as icon_path:
            root.iconbitmap(str(icon_path))


def fix_window_after_its_creation(window):
    """Perform some quality of life things after creating a window (root or Toplevel)"""
    window.attributes("-topmost", True)
    window.focus_set()
    window.after(100, lambda: window.attributes("-topmost", False))  # 100 ms


def maximise_widgets(window):
    """Maximises all widgets to look good in fullscreen"""
    # fix the grid to fill the window
    num_rows = window.grid_size()[1]  # maximise rows
    for r in range(num_rows):
        window.grid_rowconfigure(r, weight=1)
    num_cols = window.grid_size()[0]  # maximise cols
    for c in range(num_cols):
        window.grid_columnconfigure(c, weight=1)


# ..............................  change widget states  ................................


def change_widget_state_based_on_checkbox(cfg, key_to_check, widget_to_change):
    """Change the state of a widget based on state of another widget."""
    if cfg[key_to_check].get() is True:
        widget_to_change.configure(state="normal")
    elif cfg[key_to_check].get() is False:
        widget_to_change.configure(state="disabled")


def create_folder_icon():
    folder_icon = Image.open(
        resources.files("autogaita.resources").joinpath("folder.png")
    )
    return CTkImage(light_image=folder_icon, dark_image=folder_icon, size=(20, 20))



from autogaita.resources.utils import try_to_run_gaita
import os


# main function
def sleap_singlerun():
    """
    Batchrun script to run AutoGaitA SLEAP for a single dataset.
    folderinfo & cfg dictionaries must be configured as explained in our documentation. (note that each value of these corresponds to some object in the AutoGaitA Simi GUI)
    """
    # folderinfo
    folderinfo = {}
    folderinfo["root_dir"] = (
        "/Users/mahan/sciebo/Research/AutoGaitA/SLEAP/Human/Analysis/Old/"
    )
    folderinfo["results_dir"] = ""
    folderinfo["sctable_filename"] = "AnnotationTableOldSLEAP"
    folderinfo["data_string"] = ""
    folderinfo["beam_string"] = ""
    folderinfo["premouse_string"] = "Human"
    folderinfo["postmouse_string"] = "_Run"
    folderinfo["prerun_string"] = "Run"
    folderinfo["postrun_string"] = "_SLEAP"

    # cfg
    cfg = {}
    # 25 Hz sampling rate for SLEAP's example fly dataset (3000 frames @ 2 minutes)
    # https://github.com/talmolab/sleap/tree/main/docs/notebooks/analysis_example
    cfg["sampling_rate"] = 100
    cfg["subtract_beam"] = False
    cfg["dont_show_plots"] = True
    cfg["convert_to_mm"] = False
    cfg["pixel_to_mm_ratio"] = 1
    cfg["x_sc_broken_threshold"] = 200  # optional cfg
    cfg["y_sc_broken_threshold"] = 50
    cfg["x_acceleration"] = True
    cfg["angular_acceleration"] = True
    cfg["save_to_xls"] = True
    cfg["bin_num"] = 25
    cfg["plot_SE"] = True
    cfg["standardise_y_at_SC_level"] = False
    cfg["standardise_y_to_a_joint"] = False
    cfg["y_standardisation_joint"] = ["Midfoot"]
    cfg["plot_joint_number"] = 7
    cfg["color_palette"] = "viridis"
    cfg["legend_outside"] = True
    cfg["invert_y_axis"] = True
    cfg["flip_gait_direction"] = True
    cfg["analyse_average_x"] = False
    cfg["standardise_x_coordinates"] = False
    cfg["x_standardisation_joint"] = ["Midfoot"]
    cfg["coordinate_standardisation_xls"] = ""
    cfg["hind_joints"] = ["Midfoot", "Ankle", "Knee", "Hip"]
    cfg["fore_joints"] = []
    cfg["beam_col_left"] = []  # list of len == 1
    cfg["beam_col_right"] = []
    cfg["beam_hind_jointadd"] = []
    cfg["beam_fore_jointadd"] = []
    cfg["angles"] = {
        "name": ["Ankle", "Knee"],
        "lower_joint": ["Midfoot", "Ankle"],
        "upper_joint": ["Knee", "Hip"],
    }

    # info
    info = {}
    info["mouse_num"] = 32
    info["run_num"] = 1
    info["name"] = "ID " + str(info["mouse_num"]) + " - Run " + str(info["run_num"])
    if folderinfo["results_dir"]:
        info["results_dir"] = os.path.join(folderinfo["results_dir"], info["name"])
    else:
        info["results_dir"] = os.path.join(
            folderinfo["root_dir"], "Results", info["name"]
        )
    # run
    try_to_run_gaita("SLEAP", info, folderinfo, cfg, False)


# %% what happens if we just hit run
if __name__ == "__main__":
    sleap_singlerun()
